/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * Copyright (C) 2004 Sorokin R.P. (sorokinru@iias.spb.su)
 * and Object-oriented Geoinformatics Systems Laboratory (http://niggis.iias.spb.su)
 * All Rights Reserved.
 */

package ru.igis.omtab.gui;

import javax.swing.*;
import java.awt.event.*;
import com.bbn.openmap.*;
import com.bbn.openmap.omGraphics.OMGraphic;
import com.bbn.openmap.omGraphics.OMGraphicList;
import com.bbn.openmap.proj.*;
import com.bbn.openmap.proj.coords.LatLonPoint;
import com.bbn.openmap.util.*;
import com.bbn.openmap.event.*;
import ru.igis.omtab.*;
/**
 * Remake of original Openmap Distance Mouse Mode
 * to display coordinates in gegrees and minutes
 * and distances in Nautical Miles
 * @author ru
 */
public class RuDistanceMouseMode extends DistanceMouseMode {
    
    /**
	 * Comment for <code>serialVersionUID</code>
	 */
	private static final long serialVersionUID = 1L;

	// The unit type, default mile
    private Length ruUnit = Length.NM;

    // Flag to display the azimuth angle. Default true
    boolean ruShowAngle = true;
    
    private RuNA runa;
    private RuPopupMenu popup;

    /** Creates a new instance of RuDistanceMouseMode */
    public RuDistanceMouseMode() {
        super();
    }
    
    /**
     * Draw a rubberband line and circle as the mouse is moved.
     * Calculate distance and azimuth angle as the mouse moves.
     * Display distance and azimuth angle in on the infoDelegator.
     * @param e mouse event.
     */
    public void mouseMoved(MouseEvent e) {
        // set the map bean
        MapBean theMap = (MapBean)(e.getSource());
        if (theMap instanceof MapBean) {
            // set description of MapOb as ToolTip
            OMGraphicList mapObs = runa.getMapObs();
            OMGraphic closest = mapObs.findClosest(e.getX(), e.getY(), 2);
            if (closest != null) 
            	theMap.setToolTipText(closest.getDescription());
            else
            	theMap.setToolTipText(null);
            // only when the mouse has already been pressed
            if (mousePressed) {
                double lat1, lat2, long1, long2;
                // erase the old line and circle first
                paintRubberband(rPoint1, rPoint2);
                // get the current mouse location in latlon
                rPoint2 = theMap.getProjection().inverse(e.getPoint());
                // paint the new line and circle up to the current mouse location
                paintRubberband(rPoint1, rPoint2);

                if (infoDelegator != null) {
                    Debug.message("mousemodedetail", 
                                  "DistanceMouseMode: firing mouse location");
                    // lat, lon of anchor point
                    lat1  = rPoint1.getY();
                    long1 = rPoint1.getX();
                    // lat, lon of current mouse position
                    lat2  = rPoint2.getY();
                    long2 = rPoint2.getX();
                    // calculate great circle distance in nm
//                     distance = getGreatCircleDist(lat1, long1, 
//                                                   lat2, long2, Length.NM);
                    distance = GreatCircle.sphericalDistance(ProjMath.degToRad(lat1),
                                                                      ProjMath.degToRad(long1),
                                                                      ProjMath.degToRad(lat2),
                                                                      ProjMath.degToRad(long2));
                    
                    
                    // calculate azimuth angle dec deg
                    double azimuth = getSphericalAzimuth(lat1, long1, lat2, long2);
                    // convert total distance into all distance units
//                     String distNM   = df.format(totalDistance+distance);
                    double tmpDistance = totalDistance+distance;
                    String infoLine = createDistanceInformationLine(rPoint2,tmpDistance,azimuth);
                    // setup the info event
                    InfoDisplayEvent info = new InfoDisplayEvent(this, infoLine, InformationDelegator.COORDINATE_INFO_LINE);
                    // ask the infoDelegator to display the info
                    infoDelegator.requestInfoLine(info);
                }
            } else {
                fireMouseLocation(e);
            }
        }
    }
    
    protected String createDistanceInformationLine(LatLonPoint llp, double distance, double azimuth) {
        // setup the distance info to be displayed
        String unitInfo = null;
        // what unit is asked for
        if (ruUnit == null) {
            unitInfo = df.format(Length.NM.fromRadians((double)distance)) +
                Length.NM.getAbbr() + ",  " +
                df.format(Length.KM.fromRadians((double)distance)) +
                Length.KM.getAbbr() + ",  " +
                df.format(Length.MILE.fromRadians((double)distance)) +
                Length.MILE.getAbbr() + "  ";
        } else {
            unitInfo = ruUnit.fromRadians(distance) + " " + ruUnit.getAbbr();
        }

        // add the mouse lat, lon
        String infoLine = "Lat( " + MapOb.getDegMin(llp.getLatitude()) +
            " ), Lon( " + MapOb.getDegMin(llp.getLongitude())  + " ), distance( ";

        // add the units
        infoLine = infoLine + unitInfo + " )";
        // add the azimuth angle if need be
        if(azimuth<0d)
            azimuth += 360d;
        if (ruShowAngle) {
            infoLine = infoLine + ", bearings( " + df.format(azimuth) + " )";
        }
        return infoLine;
    }

    /**
     * Sends the mouse event location, x/y and lat/lon, to the
     * InformationDelegator.
     */
    public void fireMouseLocation(MouseEvent e) {
        int x = e.getX(); 
        int y = e.getY();
        LatLonPoint llp = null;
        Debug.message("mousemodedetail", "CoordMouseMode: firing mouse location");
        
        if (infoDelegator != null) {
            if (e.getSource() instanceof MapBean) {
                llp = ((MapBean)e.getSource()).getProjection().inverse(x, y);
            }   
            String infoLine;
            infoLine = createCoordinateInformationLine(x, y, llp);
            
            // setup the info event
            InfoDisplayEvent info = new InfoDisplayEvent(this, infoLine, InformationDelegator.COORDINATE_INFO_LINE);
            // ask the infoDelegator to display the info
            infoDelegator.requestInfoLine(info);
        }
    }

    /**
     * Method to create the information string reflecting information
     * at the LatLonPoint provided.  By default, will return a string
     * for the x and y, and the lat/lon.  If llp is null, just the x,
     * y will be returned.  This method can be changed, or overridden
     * to change what kind of coordinates (UTM, DMS, MGRS) are
     * reflected here.
     */
    protected String createCoordinateInformationLine(int x, int y, LatLonPoint llp) {
        if (llp != null) {
            return "Lat( " + 
                MapOb.getDegMin(llp.getLatitude()) +
                " ), Lon( " + MapOb.getDegMin(llp.getLongitude())  + 
                " ) - x( " + x + " ), y( " + y + " )";
        } else {
            return "x( " + x + " ), y( " + y + " )";
        }
    }
    
    public Icon getGUIIcon(){
        return new ImageIcon(getClass().getResource("/ru/igis/omtab/resources/Distance.gif"));
    }
    /**
     * Process a mouse pressed event. Add the mouse location to the
     * segment vector. Calculate the cumulative total distance.
     * 
     * @param e mouse event.
     */
	public void mousePressed(MouseEvent e) {
        // get MapBean object
    	MapBean theMap = (MapBean)(e.getSource());
        int button = e.getButton();
        if (button == MouseEvent.BUTTON3){
            OMGraphicList mapObs = runa.getMapObs();
            OMGraphic closest = mapObs.findClosest(e.getX(), e.getY(), 2);
       		if(closest != null)
       			showPoupMenu(e, closest, theMap);
            return;
        }
        e.getComponent().requestFocus();
        if (e.getSource() instanceof MapBean) {
            // mouse has now been pressed
            mousePressed = true;
            // erase the old circle if any
            eraseCircle();
            // anchor the new first point of the line
            rPoint1 = theMap.getProjection().inverse(e.getPoint());
            // ensure the second point is not yet set.
            rPoint2 = null;
            // add the anchor point to the list of line segments
            segments.addElement(rPoint1);
            // add the distance to the total distance
            totalDistance = totalDistance + distance;
        }
    }
    
    private void showPoupMenu(MouseEvent e, OMGraphic omg, MapBean map){
    	if(popup != null){
    		popup.event(e, (MapOb)omg);
    		popup.show(map, e.getX(), e.getY());
        }
    }
    /**
     * Standard OpenMap interface LightMapHandlerChild method.
     * Search for RuA and RuPopupMenu
     * @param obj - serched object
     */    
    public void findAndInit(Object obj) {
        if (obj instanceof RuNA) {
            runa = (RuNA)obj;
        } else if (obj instanceof RuPopupMenu) {
            popup = (RuPopupMenu)obj;
        } else if (obj instanceof InformationDelegator) {
            setInfoDelegator((InformationDelegator) obj);
        }
    }

    /**
     * Standard OpenMap interface LightMapHandlerChild method.
     * Set to null links 
     * in Clock object
     * @param obj - searched object
     */    
    public void findAndUndo(Object obj) {
        if (obj instanceof RuNA) {
            runa = null;
        } else if (obj instanceof RuPopupMenu) {
            popup = null;
        } else if (obj instanceof InformationDelegator) {
            if (getInfoDelegator() == (InformationDelegator) obj) {
                setInfoDelegator(null);
            }
        }
    }
}
